
#pragma Ospace

#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"

#include "procstate.h"
#include "datetime.h"

#include "glib/glib.h"

#include "fat2.h"
#include "shell.h"

#include "skin.h"
#include "component.h"
#include "sndeff.h"
#include "lang.h"
#include "strpcm.h"
#include "rect.h"

static TProcState *pCurrentProcState;

#define ClientHeight (192+128+48)

// -----------------------------

static CglCanvas *pScreenBack;
static s32 ScreenPosY,ShowPosY;

static void ScrollBar_SetPosY(s32 PosY)
{
  ScreenPosY=PosY;
  if(ScreenPosY<0) ScreenPosY=0;
  if((ClientHeight-ScreenHeight)<ScreenPosY) ScreenPosY=ClientHeight-ScreenHeight;
}

// -----------------------------

enum ECompLabels {ECLS_TitleLbl,ECLS_TopScrModeTitleLbl,ECLS_LineSpaceTitleLbl,ECLS_FontSizeTitleLbl,ECLS_CPTitleLbl,ECLS_DetectCharCodeTitleLbl,ECLSCount};
#define CompLabelsCount (ECLSCount)
static TComponentLabel CompLabels[CompLabelsCount];

enum ECompChecks {ECCS_TopScrModeLightOffRadio,ECCS_TopScrModeTextRadio,ECCS_TopScrModeClockRadio,
                  ECCS_LineSpaceSmallRadio,ECCS_LineSpaceMiddleRadio,ECCS_LineSpaceLargeRadio,
                  ECCS_ScreenSaver_BlackModeChk,
                  ECCS_FontSizeSmallRadio,ECCS_FontSizeMiddleRadio,ECCS_FontSizeLargeRadio,
                  ECCS_CP437Radio,ECCS_CP850Radio,ECCS_CP1252Radio,
                  ECCS_DetectCharCode_ANSIChk,ECCS_DetectCharCode_SJISChk,ECCS_DetectCharCode_UTF16BEChk,ECCS_DetectCharCode_UTF16LEChk,ECCS_DetectCharCode_UTF8Chk,
                  ECCSCount};
#define CompChecksCount (ECCSCount)
static TComponentCheck CompChecks[CompChecksCount];

enum ECompButtons {ECBS_CancelBtn,ECBS_OkBtn,ECBSCount};
#define CompButtonsCount (ECBSCount)
static TComponentButton CompButtons[CompButtonsCount];

static void Setting_Redraw_TransOnly(void)
{
  s32 posy=ShowPosY;
  
  if(posy<ScreenHeight){
    CglB15 *pb15=Custom_GetSkin(ECS_TopMsg);
    pb15->BitBlt(pScreenSub->pCanvas,0,0,ScreenWidth,ScreenHeight-posy,0,0);
  }
  
  pScreenBack->BitBlt(pScreenSub->pCanvas,0,ScreenHeight-posy,ScreenWidth,posy,0,0,false);
  
  pScreenBack->BitBlt(pScreenMain->pBackCanvas,0,0,ScreenWidth,ScreenHeight,0,posy,false);
  
  {
    CglCanvas *pcan=pScreenMain->pBackCanvas;
    const u16 FrameColor=ColorTable.Component.SoftwareScrollBar_Frame;
    const u16 BGColor=ColorTable.Component.SoftwareScrollBar_BG;
    
    s32 w=10,h=(ScreenHeight*ScreenHeight)/ClientHeight;
    s32 x=ScreenWidth-2-w,y=posy/2;
    pcan->SetColor(FrameColor);
    pcan->DrawBox(x,y,w,h);
    
    x+=1; y+=1; w-=2; h-=2;
    u16 *pbuf=pcan->GetScanLine(y);
    pbuf+=x;
    u32 bufsize=pcan->GetWidth()-w;
    for(s32 py=0;py<h;py++){
      for(s32 px=0;px<w;px++){
        u16 col=*pbuf;
        u16 mask=RGB15(30,30,30);
        col=(col&mask)>>1;
        *pbuf++=(BGColor+col)|BIT15;
      }
      pbuf+=bufsize;
    }
  }
  
  ScreenMain_Flip_ProcFadeEffect();
}

static void Setting_Redraw(void)
{
  {
    CglB15 *pb15=Custom_GetSkin(ECS_BG);
    pb15->pCanvas->BitBltFullBeta(pScreenBack);
  }
  
  for(u32 idx=0;idx<CompChecksCount;idx++){
    TComponentCheck *pcc=&CompChecks[idx];
    bool *pchk=(bool*)pcc->UserData;
    if(pchk!=NULL) pcc->Checked=*pchk;
  }
  
  {
    CompChecks[ECCS_LineSpaceSmallRadio].Checked=false;
    CompChecks[ECCS_LineSpaceMiddleRadio].Checked=false;
    CompChecks[ECCS_LineSpaceLargeRadio].Checked=false;
    
    switch(pCurrentProcState->Text.LineSpace){
      case ETLS_Small: CompChecks[ECCS_LineSpaceSmallRadio].Checked=true; break;
      case ETLS_Middle: CompChecks[ECCS_LineSpaceMiddleRadio].Checked=true; break;
      case ETLS_Large: CompChecks[ECCS_LineSpaceLargeRadio].Checked=true; break;
      default: break;
    }
  }
  
  {
    CompChecks[ECCS_TopScrModeTextRadio].Checked=false;
    CompChecks[ECCS_TopScrModeClockRadio].Checked=false;
    CompChecks[ECCS_TopScrModeLightOffRadio].Checked=false;
    
    switch(pCurrentProcState->Text.TopScrMode){
      case ETTSM_Text: CompChecks[ECCS_TopScrModeTextRadio].Checked=true; break;
      case ETTSM_Clock: CompChecks[ECCS_TopScrModeClockRadio].Checked=true; break;
      case ETTSM_LightOff: CompChecks[ECCS_TopScrModeLightOffRadio].Checked=true; break;
      default: break;
    }
  }
  
  {
    CompChecks[ECCS_FontSizeSmallRadio].Checked=false;
    CompChecks[ECCS_FontSizeMiddleRadio].Checked=false;
    CompChecks[ECCS_FontSizeLargeRadio].Checked=false;
    
    switch(pCurrentProcState->Text.FontSize){
      case Text_FontSize_Small: CompChecks[ECCS_FontSizeSmallRadio].Checked=true; break;
      case Text_FontSize_Middle: CompChecks[ECCS_FontSizeMiddleRadio].Checked=true; break;
      case Text_FontSize_Large: CompChecks[ECCS_FontSizeLargeRadio].Checked=true; break;
      default: break;
    }
  }
  
  {
    CompChecks[ECCS_CP437Radio].Checked=false;
    CompChecks[ECCS_CP850Radio].Checked=false;
    CompChecks[ECCS_CP1252Radio].Checked=false;
    
    switch(pCurrentProcState->Text.DefaultCodePage){
      case 437: CompChecks[ECCS_CP437Radio].Checked=true; break;
      case 850: CompChecks[ECCS_CP850Radio].Checked=true; break;
      case 1252: CompChecks[ECCS_CP1252Radio].Checked=true; break;
      default: break;
    }
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Draw(&CompLabels[idx]);
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Draw(&CompChecks[idx]);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Draw(&CompButtons[idx]);
  }
  
  Setting_Redraw_TransOnly();
}

static void CB_TopScrModeLightOffRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.TopScrMode=ETTSM_LightOff;
}

static void CB_TopScrModeTextRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.TopScrMode=ETTSM_Text;
}

static void CB_TopScrModeClockRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.TopScrMode=ETTSM_Clock;
}

static void CB_LineSpaceSmallRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.LineSpace=ETLS_Small;
}

static void CB_LineSpaceMiddleRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.LineSpace=ETLS_Middle;
}

static void CB_LineSpaceLargeRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.LineSpace=ETLS_Large;
}

static void CB_FontSizeSmallRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.FontSize=Text_FontSize_Small;
}

static void CB_FontSizeMiddleRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.FontSize=Text_FontSize_Middle;
}

static void CB_FontSizeLargeRadio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.FontSize=Text_FontSize_Large;
}

static void CB_CP437Radio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.DefaultCodePage=437;
}

static void CB_CP850Radio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.DefaultCodePage=850;
}

static void CB_CP1252Radio_Click(void *pComponentCheck)
{
  Sound_Start(WAVFN_Click);
  pCurrentProcState->Text.DefaultCodePage=1252;
}

static void CB_CancelBtn_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  SetNextProc(ENP_TextView,EPFE_LeftToRight);
}

static void CB_OkBtn_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  
  ProcState=*pCurrentProcState;
  ProcState_RequestSave=true;
  ProcState_Save();
  
  SetNextProc(ENP_TextView,EPFE_LeftToRight);
}

static void CB_CheckBox_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  TComponentCheck *pcc=(TComponentCheck*)pComponentButton;
  bool *pchk=(bool*)pcc->UserData;
  if(pchk!=NULL) *pchk=!*pchk;
}

static void CompsInit(void)
{
  CglCanvas *pcan=pScreenBack;
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Init(&CompLabels[idx],pcan);
    TComponentLabel *pcl=&CompLabels[idx];
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Init(&CompChecks[idx],pcan);
    TComponentCheck *pcc=&CompChecks[idx];
    pcc->CallBack_Click=CB_CheckBox_Click;
    pcc->UserData=NULL;
    pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_CheckOn);
    pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_CheckOff);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Init(&CompButtons[idx],pcan);
    TComponentButton *pcb=&CompButtons[idx];
  }
  
  // -----------------
  
  u32 TextHeight=glCanvasTextHeight;
  
  u32 x=0,y=0,w=0,h=0;
  
  h=TextHeight;
  
  // -----------------
  
  x=8;
  y+=8;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_TitleLbl];
    pcl->pMsgUTF8=Lang_GetUTF8("TCS_Title");
    pcl->Rect=CreateRect(x+24,y,ScreenWidth-x,h);
  }
  y+=12*2;

  // -----------------
  
  u32 chkw=24,chkh=14;
  
  u32 PaddingY=8;
  
  y+=PaddingY;
  
  TProcState *pcps=pCurrentProcState;
  
  // -----------------
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_TopScrModeTitleLbl];
    pcl->pMsgUTF8=Lang_GetUTF8("TCS_TopScrModeTitle");
    pcl->Rect=CreateRect(x,y,ScreenWidth-x,h);
  }
  y+=TextHeight+PaddingY;
    
  {
    u32 rx=x;
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_TopScrModeLightOffRadio];
      pcc->CallBack_Click=CB_TopScrModeLightOffRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_TopScrModeLightOff");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_TopScrModeTextRadio];
      pcc->CallBack_Click=CB_TopScrModeTextRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_TopScrModeText");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_TopScrModeClockRadio];
      pcc->CallBack_Click=CB_TopScrModeClockRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_TopScrModeClock");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
  }
  
  y+=TextHeight*2;
  
  // -----------------
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_ScreenSaver_BlackModeChk];
    pcc->UserData=(u32)&pcps->Text.ScreenSaver_BlackMode;
    pcc->pMsgUTF8=Lang_GetUTF8("TCS_ScreenSaver_BlackMode");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  y+=TextHeight*2;
  
  // -----------------
  
  {
    u32 rx=x;
    
    {
      TComponentLabel *pcl=&CompLabels[ECLS_LineSpaceTitleLbl];
      pcl->pMsgUTF8=Lang_GetUTF8("TCS_LineSpaceTitle");
      pcl->Rect=CreateRect(x,y,ScreenWidth-x,h);
      rx+=ComponentLabel_GetWidth(pcl);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_LineSpaceSmallRadio];
      pcc->CallBack_Click=CB_LineSpaceSmallRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_LineSpaceSmall");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_LineSpaceMiddleRadio];
      pcc->CallBack_Click=CB_LineSpaceMiddleRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_LineSpaceMiddle");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_LineSpaceLargeRadio];
      pcc->CallBack_Click=CB_LineSpaceLargeRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_LineSpaceLarge");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
  }
  
  y+=TextHeight*2;
  y+=PaddingY;
  
  // -----------------
  
  {
    u32 rx=x;
    
    {
      TComponentLabel *pcl=&CompLabels[ECLS_FontSizeTitleLbl];
      pcl->pMsgUTF8=Lang_GetUTF8("TCS_FontSizeTitle");
      pcl->Rect=CreateRect(rx,y,ScreenWidth-x,h);
      rx+=ComponentLabel_GetWidth(pcl);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_FontSizeSmallRadio];
      pcc->CallBack_Click=CB_FontSizeSmallRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_FontSizeSmall");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_FontSizeMiddleRadio];
      pcc->CallBack_Click=CB_FontSizeMiddleRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_FontSizeMiddle");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_FontSizeLargeRadio];
      pcc->CallBack_Click=CB_FontSizeLargeRadio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_FontSizeLarge");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
  }
  
  y+=TextHeight*2;
  
  // -----------------
  
  y+=8;
  
  // -----------------
  
  {
    u32 rx=x;
    
    {
      TComponentLabel *pcl=&CompLabels[ECLS_CPTitleLbl];
      pcl->pMsgUTF8=Lang_GetUTF8("TCS_CPTitle");
      pcl->Rect=CreateRect(rx,y,ScreenWidth-x,h);
      rx+=ComponentLabel_GetWidth(pcl);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_CP437Radio];
      pcc->CallBack_Click=CB_CP437Radio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_CP437");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_CP850Radio];
      pcc->CallBack_Click=CB_CP850Radio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_CP850");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
    
    {
      TComponentCheck *pcc=&CompChecks[ECCS_CP1252Radio];
      pcc->CallBack_Click=CB_CP1252Radio_Click;
      pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_RadioOn);
      pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_RadioOff);
      pcc->pMsgUTF8=Lang_GetUTF8("TCS_CP1252");
      pcc->Rect=CreateRect(rx,y,chkw,chkh);
      rx+=ComponentCheck_GetWidth(pcc);
    }
  }
    
  y+=TextHeight*2;
  
  // -----------------
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_DetectCharCodeTitleLbl];
    pcl->pMsgUTF8=Lang_GetUTF8("TCS_DetectCharCodeTitle");
    pcl->Rect=CreateRect(x,y,ScreenWidth-x,h);
  }
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_DetectCharCode_ANSIChk];
    pcc->UserData=(u32)&pcps->Text.DetectCharCode_ANSI;
    pcc->pMsgUTF8=Lang_GetUTF8("TCS_DetectCharCode_ANSI");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_DetectCharCode_SJISChk];
    pcc->UserData=(u32)&pcps->Text.DetectCharCode_SJIS;
    pcc->pMsgUTF8=Lang_GetUTF8("TCS_DetectCharCode_SJIS");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_DetectCharCode_UTF16BEChk];
    pcc->UserData=(u32)&pcps->Text.DetectCharCode_UTF16BE;
    pcc->pMsgUTF8=Lang_GetUTF8("TCS_DetectCharCode_UTF16BE");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_DetectCharCode_UTF16LEChk];
    pcc->UserData=(u32)&pcps->Text.DetectCharCode_UTF16LE;
    pcc->pMsgUTF8=Lang_GetUTF8("TCS_DetectCharCode_UTF16LE");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  {
    TComponentCheck *pcc=&CompChecks[ECCS_DetectCharCode_UTF8Chk];
    pcc->UserData=(u32)&pcps->Text.DetectCharCode_UTF8;
    pcc->pMsgUTF8=Lang_GetUTF8("TCS_DetectCharCode_UTF8");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  y+=TextHeight*2;
  
  // -----------------
  
  u32 padx=24;
  
  w=96;
  h=24;
  x=padx;
  y=ClientHeight-h-PaddingY;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_CancelBtn];
    pcb->CallBack_Click=CB_CancelBtn_Click;
    pcb->pIcon=ComponentAlpha_GetSkin(ECSA_Cancel);
    pcb->pMsgUTF8=Lang_GetUTF8("CS_Cancel");
    pcb->Rect=CreateRect(x,y,w,h);
  }
  
  w=72;
  x=ScreenWidth-padx-w;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_OkBtn];
    pcb->CallBack_Click=CB_OkBtn_Click;
    pcb->pIcon=ComponentAlpha_GetSkin(ECSA_Ok);
    pcb->pMsgUTF8=Lang_GetUTF8("CS_Ok");
    pcb->Rect=CreateRect(x,y,w,h);
  }
}

// -----------------------------

static void CB_KeyDown(u32 Keys)
{
}

static void CB_KeyPress(u32 VsyncCount,u32 Keys)
{
  if((Keys&KEY_B)!=0){
    CB_CancelBtn_Click(NULL);
  }
  
  if((Keys&KEY_A)!=0){
    CB_OkBtn_Click(NULL);
  }
  
  if((Keys&(KEY_UP|KEY_DOWN|KEY_LEFT|KEY_RIGHT))!=0){
    s32 v=0;
    
    if((Keys&(KEY_UP|KEY_LEFT))!=0) v=-1;
    if((Keys&(KEY_DOWN|KEY_RIGHT))!=0) v=1;
    
    ScrollBar_SetPosY(ScreenPosY+(32*v));
  }
}

static void CB_KeyUp(u32 Keys)
{
}

static bool scrmf;
static s32 scrmy;

static bool deskmf;
static TComponentButton *pPressingButton;

static void CB_MouseDown(s32 x,s32 y)
{
  if((ScreenWidth-24)<=x){
    scrmf=true;
    scrmy=y;
    s32 h=(ScreenHeight*ScreenHeight)/ClientHeight;
    ScrollBar_SetPosY(((scrmy*ClientHeight)/ScreenHeight)-(h/2));
    ShowPosY=ScreenPosY;
    Setting_Redraw_TransOnly();
    return;
  }
  
  deskmf=false;
  
  y+=ScreenPosY;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
      pPressingButton=pcb;
      pcb->Pressing=true;
      ComponentButton_Draw(pcb);
      Setting_Redraw();
      deskmf=true;
      return;
    }
  }
  
  deskmf=true;
}

static void CB_MouseMove(s32 x,s32 y)
{
  if(scrmf==true){
    scrmy=y;
    s32 h=(ScreenHeight*ScreenHeight)/ClientHeight;
    ScrollBar_SetPosY(((scrmy*ClientHeight)/ScreenHeight)-(h/2));
    ShowPosY=ScreenPosY;
    Setting_Redraw_TransOnly();
    return;
  }
  
  if(deskmf==false) return;
  
  y+=ScreenPosY;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(pcb==pPressingButton){
      if(ComponentButton_GetIndexFromPos(pcb,x,y)==-1){
        if(pcb->Pressing==true){
          pcb->Pressing=false;
          ComponentButton_Draw(pcb);
          Setting_Redraw();
        }
        }else{
        if(pcb->Pressing==false){
          pcb->Pressing=true;
          ComponentButton_Draw(pcb);
          Setting_Redraw();
        }
      }
    }
  }
}

static void CB_MouseUp(s32 x,s32 y)
{
  if(scrmf==true){
    scrmf=false;
    scrmy=y;
    s32 h=(ScreenHeight*ScreenHeight)/ClientHeight;
    ScrollBar_SetPosY(((scrmy*ClientHeight)/ScreenHeight)-(h/2));
    ShowPosY=ScreenPosY;
    Setting_Redraw_TransOnly();
    return;
  }
  
  if(deskmf==false) return;
  
  deskmf=false;
  
  y+=ScreenPosY;
  
  if(pPressingButton!=NULL){
    pPressingButton->Pressing=false;
    ComponentButton_Draw(pPressingButton);
    
    for(u32 idx=0;idx<CompButtonsCount;idx++){
      TComponentButton *pcb=&CompButtons[idx];
      if(pcb==pPressingButton){
        if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
          ComponentButton_MouseUp(&CompButtons[idx],x,y);
        }
      }
    }
    
    Setting_Redraw();
    pPressingButton=NULL;
    return;
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    if(ComponentLabel_MouseUp(&CompLabels[idx],x,y)==true) Setting_Redraw();
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    if(ComponentCheck_MouseUp(&CompChecks[idx],x,y)==true) Setting_Redraw();
  }
  
}

static void CB_Start(void)
{
  Sound_Start(WAVFN_Open);
  
  scrmf=false;
  scrmy=0;
  
  deskmf=false;
  pPressingButton=NULL;
  
  pScreenBack=new CglCanvas(NULL,ScreenWidth,ScreenHeight*2,pf15bit);
  ScrollBar_SetPosY(0);
  ShowPosY=ScreenPosY;
  
  pCurrentProcState=(TProcState*)safemalloc(sizeof(TProcState));
  if(pCurrentProcState==NULL){
    _consolePrintf("Fatal error: pCurrentProcState memory overflow.\n");
    ShowLogHalt();
  }
  *pCurrentProcState=ProcState;
  
  CompsInit();
  
  Setting_Redraw();
}

static void CB_VsyncUpdate(u32 VsyncCount)
{
  if(ShowPosY!=ScreenPosY){
    VsyncCount*=4;
    while(VsyncCount!=0){
      VsyncCount--;
      if(ScreenPosY!=ShowPosY){
        if(ScreenPosY<ShowPosY){
          ShowPosY--;
          }else{
          ShowPosY++;
        }
      }
    }
    Setting_Redraw_TransOnly();
  }
}

static void CB_End(void)
{
  if(pCurrentProcState!=NULL){
    safefree(pCurrentProcState); pCurrentProcState=NULL;
  }
  
  if(pScreenBack!=NULL){
    delete pScreenBack; pScreenBack=NULL;
  }
}

void ProcTextCustom_SetCallBack(void)
{
  CallBack_Start=CB_Start;
  CallBack_VsyncUpdate=CB_VsyncUpdate;
  CallBack_End=CB_End;
  CallBack_KeyDown=CB_KeyDown;
  CallBack_KeyPress=CB_KeyPress;
  CallBack_KeyUp=CB_KeyUp;
  CallBack_MouseDown=CB_MouseDown;
  CallBack_MouseMove=CB_MouseMove;
  CallBack_MouseUp=CB_MouseUp;
}

