
#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"

#include "procstate.h"
#include "datetime.h"

#include "glib/glib.h"

#include "fat2.h"
#include "shell.h"

#include "skin.h"
#include "component.h"
#include "sndeff.h"

enum ECompLabels {ECLSCount};
#define CompLabelsCount (ECLSCount)
static TComponentLabel CompLabels[CompLabelsCount];

enum ECompChecks {ECCS_BootCheckDiskChk,ECCS_ClickSoundChk,
                  ECCS_FileListMode_SingleRadio,ECCS_FileListMode_DoubleRadio,
                  ECCS_SkipSetupChk,ECCSCount};
#define CompChecksCount (ECCSCount)
static TComponentCheck CompChecks[CompChecksCount];

enum ECompButtons {ECBS_OKBtn,ECBSCount};
#define CompButtonsCount (ECBSCount)
static TComponentButton CompButtons[CompButtonsCount];

static void Setting_Redraw(void)
{
  {
    CglCanvas *pcan=pScreenMain->pBackCanvas;
    
    CglB15 *pbg=Setup_GetSkin(ESS_BG_Bottom);
    pbg->pCanvas->BitBltFullBeta(pcan);
  }
  
  TProcState_System *psys=&ProcState.System;
  
  CompChecks[ECCS_BootCheckDiskChk].Checked=psys->BootCheckDisk;
  CompChecks[ECCS_ClickSoundChk].Checked=psys->ClickSound;
  CompChecks[ECCS_SkipSetupChk].Checked=psys->SkipSetup;
  
  {
    TProcState_FileList *pfl=&ProcState.FileList;
    
    CompChecks[ECCS_FileListMode_SingleRadio].Checked=false;
    CompChecks[ECCS_FileListMode_DoubleRadio].Checked=false;
    
    switch(pfl->Mode){
      case EPSFLM_Single: CompChecks[ECCS_FileListMode_SingleRadio].Checked=true; break;
      case EPSFLM_Double: CompChecks[ECCS_FileListMode_DoubleRadio].Checked=true; break;
      default: break;
    }
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Draw(&CompLabels[idx]);
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Draw(&CompChecks[idx]);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Draw(&CompButtons[idx]);
  }
  
  ScreenMain_Flip_ProcFadeEffect();
}

static void CB_BootCheckDiskChk_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  TProcState_System *psys=&ProcState.System;
  psys->BootCheckDisk=!psys->BootCheckDisk;
  SetProcFadeEffect(EPFE_FastCrossFade);
}

static void CB_ClickSoundChk_Click(void *pComponentButton)
{
  TProcState_System *psys=&ProcState.System;
  psys->ClickSound=!psys->ClickSound;
  Sound_Start(WAVFN_Click);
  SetProcFadeEffect(EPFE_FastCrossFade);
}

static void CB_SkipSetupChk_Click(void *pComponentButton)
{
  TProcState_System *psys=&ProcState.System;
  psys->SkipSetup=!psys->SkipSetup;
  Sound_Start(WAVFN_Click);
  SetProcFadeEffect(EPFE_FastCrossFade);
}

static void CB_FileListMode_SingleRadio_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  TProcState_FileList *pfl=&ProcState.FileList;
  pfl->Mode=EPSFLM_Single;
  SetProcFadeEffect(EPFE_FastCrossFade);
}

static void CB_FileListMode_DoubleRadio_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  TProcState_FileList *pfl=&ProcState.FileList;
  pfl->Mode=EPSFLM_Double;
  SetProcFadeEffect(EPFE_FastCrossFade);
}

static void CB_OKBtn_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  switch(ProcState.System.LastState){
    case ELS_FileList: SetNextProc(ENP_FileList,EPFE_CrossFade); break;
    case ELS_Launch: SetNextProc(ENP_Launch,EPFE_CrossFade); break;
  }
}

static void CompsInit(void)
{
  CglCanvas *pcan=pScreenMain->pBackCanvas;
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Init(&CompLabels[idx],pcan);
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Init(&CompChecks[idx],pcan);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Init(&CompButtons[idx],pcan);
  }
  
  s32 chksize=13;
  
  s32 x,y;
  
  x=8;
  y=8;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_BootCheckDiskChk];
    pcc->CallBack_Click=CB_BootCheckDiskChk_Click;
    pcc->pOnIcon=SetupAlpha_GetSkin(ESSA_ChkOverlayOn);
    pcc->pOffIcon=NULL;
    pcc->pMsgUTF8="";
    pcc->Rect=CreateRect(x,y,ScreenWidth-x,chksize+5);
    pcc->Centering=false;
  }
  
  y+=28;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_ClickSoundChk];
    pcc->CallBack_Click=CB_ClickSoundChk_Click;
    pcc->pOnIcon=SetupAlpha_GetSkin(ESSA_ChkOverlayOn);
    pcc->pOffIcon=NULL;
    pcc->pMsgUTF8="";
    pcc->Rect=CreateRect(x,y,ScreenWidth-x,chksize+5);
    pcc->Centering=false;
  }
  
  x+=9;
  y=84;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_FileListMode_SingleRadio];
    pcc->CallBack_Click=CB_FileListMode_SingleRadio_Click;
    pcc->pOnIcon=SetupAlpha_GetSkin(ESSA_RadioOverlayOn);
    pcc->pOffIcon=NULL;
    pcc->pMsgUTF8="";
    pcc->Rect=CreateRect(x,y,ScreenWidth-x,chksize+5);
    pcc->Centering=false;
  }
  
  y+=20;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_FileListMode_DoubleRadio];
    pcc->CallBack_Click=CB_FileListMode_DoubleRadio_Click;
    pcc->pOnIcon=SetupAlpha_GetSkin(ESSA_RadioOverlayOn);
    pcc->pOffIcon=NULL;
    pcc->pMsgUTF8="";
    pcc->Rect=CreateRect(x,y,ScreenWidth-x,chksize+5);
    pcc->Centering=false;
  }
  
  x-=9;
  
  y+=36;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_SkipSetupChk];
    pcc->CallBack_Click=CB_SkipSetupChk_Click;
    pcc->pOnIcon=SetupAlpha_GetSkin(ESSA_ChkOverlayOn);
    pcc->pOffIcon=NULL;
    pcc->pMsgUTF8="";
    pcc->Rect=CreateRect(x,y,ScreenWidth-x,chksize+5);
    pcc->Centering=false;
  }
  
  x=148;
  y=160;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_OKBtn];
    pcb->CallBack_Click=CB_OKBtn_Click;
    pcb->pIcon=NULL;
    pcb->pMsgUTF8="";
    pcb->Rect=CreateRect(x,y,ScreenWidth-x,ScreenHeight-y);
    pcb->DrawFrame=false;
  }
}

// -----------------------------

static void CB_KeyDown(u32 Keys)
{
}

static void CB_KeyPress(u32 VsyncCount,u32 Keys)
{
  if( ((Keys&KEY_A)!=0) || ((Keys&KEY_B)!=0) ) CB_OKBtn_Click(NULL);
}

static void CB_KeyUp(u32 Keys)
{
}

static bool deskmf;
static TComponentButton *pPressingButton;

static void CB_MouseDown(s32 x,s32 y)
{
  deskmf=false;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
      pPressingButton=pcb;
      pcb->Pressing=true;
      ComponentButton_Draw(pcb);
      Setting_Redraw();
    }
  }
  
  deskmf=true;
}

static void CB_MouseMove(s32 x,s32 y)
{
  if(deskmf==false) return;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(pcb==pPressingButton){
      if(ComponentButton_GetIndexFromPos(pcb,x,y)==-1){
        if(pcb->Pressing==true){
          pcb->Pressing=false;
          ComponentButton_Draw(pcb);
          Setting_Redraw();
        }
        }else{
        if(pcb->Pressing==false){
          pcb->Pressing=true;
          ComponentButton_Draw(pcb);
          Setting_Redraw();
        }
      }
    }
  }
}

static void CB_MouseUp(s32 x,s32 y)
{
  if(deskmf==false) return;
  
  deskmf=false;
  
  if(pPressingButton!=NULL){
    pPressingButton->Pressing=false;
    ComponentButton_Draw(pPressingButton);
    
    for(u32 idx=0;idx<CompButtonsCount;idx++){
      TComponentButton *pcb=&CompButtons[idx];
      if(pcb==pPressingButton){
        if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
          ComponentButton_MouseUp(&CompButtons[idx],x,y);
        }
      }
    }
    
    Setting_Redraw();
    pPressingButton=NULL;
    return;
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    if(ComponentLabel_MouseUp(&CompLabels[idx],x,y)==true) Setting_Redraw();
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    if(ComponentCheck_MouseUp(&CompChecks[idx],x,y)==true) Setting_Redraw();
  }
}

static void CB_Start(void)
{
  {
    CglCanvas *pcan=pScreenSub->pCanvas;
    
    CglB15 *pbg=Setup_GetSkin(ESS_BG_Top);
    pbg->pCanvas->BitBltFullBeta(pcan);
  }
  
  deskmf=false;
  pPressingButton=NULL;
  
  CompsInit();
  
  Sound_Start(WAVFN_Open);
  
  Setting_Redraw();
}

static void CB_VsyncUpdate(u32 VsyncCount)
{
}

static void CB_End(void)
{
  ProcState_RequestSave=true;
  ProcState_Save();
}

void ProcSetup_SetCallBack(void)
{
  CallBack_Start=CB_Start;
  CallBack_VsyncUpdate=CB_VsyncUpdate;
  CallBack_End=CB_End;
  CallBack_KeyDown=CB_KeyDown;
  CallBack_KeyPress=CB_KeyPress;
  CallBack_KeyUp=CB_KeyUp;
  CallBack_MouseDown=CB_MouseDown;
  CallBack_MouseMove=CB_MouseMove;
  CallBack_MouseUp=CB_MouseUp;
}

